#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <string.h>
#include <mysql.h>
#include "connection.h"
#include "result.h"
#include "row.h"
#include "lpsql.h"

using namespace std;

#define LPSQLDEBUG //output message for debugging.
#ifdef LPSQLDEBUG
#include <time.h>  //to measure calculating time
#endif

#define READONLY

/*  lpsqldecodepath  -- decode SQL path                          *\
INPUT:
  path               : SQL path
  start_id_extension : "id" or "idx" at the first table
OUTPUT:
  table_name         : Table name list
  transition_depth   : depth of transition from noArray to Array
  id_extension       : ID extension list ("id" or "idx")
  field_name         : field name at the last table
  group_name         : group_name at the transition point
RETURN:
  maxdepth           : depth of the last table
\*                                                               */

int lpsqldecodepath(
		    char *path
		    ,char *start_id_extension
		    ,char **table_name
		    ,char **id_extension
		    ,int  &transition_depth
		    ,char *group_name
		    ,char *field_name
		    ){
  char sqlpath[1024];
  strcpy(sqlpath,path);
  char* pnext1 = sqlpath;
  char* pnext2;
  int   depth  = 0;
  int   maxdepth;
  transition_depth = -1;
  do{
    strcpy(table_name[depth],strtok_r(pnext1,"/",&pnext1));
    if(table_name[depth][0] == '('){
      transition_depth = depth;
      strcpy(group_name,&table_name[depth][1]);  
      group_name[strlen(group_name)-1]='\0'; //remove ')'
      continue;
    }
    pnext2 = table_name[depth];    
    if(strchr(pnext2,'|')){
      do{
	strcpy(table_name[depth],strtok_r(pnext2,"|",&pnext2));
	if(table_name[depth][0] == '('){
          transition_depth = depth;
	  strcpy(group_name,&table_name[depth][1]);  
	  group_name[strlen(group_name)-1]='\0'; //remove ')'
	  continue;
	}
	depth++;
	strcpy(id_extension[depth],"idx");
      }while(strlen(pnext2));
      depth--;
    }
    depth++;
    strcpy(id_extension[depth],"id");    
  }while(strlen(pnext1));
  
  maxdepth=depth-1;
  strcpy(id_extension[0],start_id_extension);
  //get field name
  pnext1=table_name[maxdepth];
  strtok_r(pnext1,".",&pnext1);
  strcpy(field_name,pnext1);

  return maxdepth;
}

//-------------------------//
//         lpsqlread
//-------------------------//
Result* lpsqlread(Connection *connection, char *path, int start_id, char *start_id_extension,int max=1000){
  Result     *result;
  Row        *row;
  char       sqlquery[2048];
  
  char from_phrase[4096]="";
  char where_phrase[4096]="";
  char order_phrase[4096]="";
  char limit_phrase[4096]="";
  char group_definition[3][4096];
  char table_name[32][128];
  char *ptable_name[32];
  char field_name[128];
  char entry[32][128];
  char id_extension[32][8];    // "id","idx"
  char *pid_extension[32];
  int  transition_depth;   // 
  int  nentry       = 0;
  int  maxdepth     = 0;
  int  depth        = 0;
  char group_name[128];

  char str[80];
  char* pnext1 = path;
  int  i;

  for(i=0;i<32;i++){
    ptable_name[i]=table_name[i];
    pid_extension[i]=id_extension[i];
  }
  maxdepth = lpsqldecodepath(path,start_id_extension
			     ,ptable_name,pid_extension,transition_depth
			     ,group_name,field_name);
#ifdef LPSQLDEBUG
  cout<<"Path\t:"<<path<<endl;
  for(depth=0;depth<=maxdepth;depth++){
    cout<<"Table\t:"<<depth<<" "<<table_name[depth]<<" "
	<<" "<<id_extension[depth]<<endl;
  }
  clock_t oldTime,newTime;
  oldTime = clock();
  cout<<"Field\t:"<<field_name<<endl<<endl;
  cout<<"----   Start following relations   ----"<<endl;
#endif
  
  //--  start following relation.  --//
  sprintf(str,"%s.%s = %d",table_name[0],id_extension[0],start_id);    
  strcat(where_phrase,str);
  
  for(depth=0;depth<=maxdepth;depth++){
#ifdef LPSQLDEBUG
    cout<<"Level-"<<depth<<": "<<table_name[depth]<<endl;
#endif
    
    //add table to FROM phrase
    if(transition_depth != depth){
      if(depth){
	strcat(from_phrase,",");
      }
      strcat(from_phrase,table_name[depth]);
    }
    
    //add relation with next table to WHERE phrase    
    if(depth!=maxdepth && transition_depth != depth+1){
      strcat(where_phrase," and ");
      sprintf(str,"%s.%s_%s = %s.%s"
	      ,table_name[depth]  , table_name[depth+1]  , id_extension[depth+1]
	      ,table_name[depth+1], id_extension[depth+1]);    
      strcat(where_phrase,str);
    }

    //Read informations at connecting node (nonArray -> Array)
    if(transition_depth==depth+1){
      //get group definition
	sprintf(sqlquery
		,"SELECT %s.%s_definition FROM %s WHERE %s;"
		,table_name[depth],group_name,from_phrase,where_phrase); 
	cout<<sqlquery<<endl;	
	result = connection->Query(sqlquery);
	result->Next();
	row = result->GetCurrentRow();
	pnext1 = row->GetField(1);
	cout<<"OK"<<endl;	
	//decode group definition
	strcpy(group_definition[0],strtok_r(pnext1,":",&pnext1)); //table list
	strcpy(group_definition[1],strtok_r(pnext1,":",&pnext1)); //condition
	strcpy(group_definition[2],pnext1);                       //order
	
	//decode menber list.    
	nentry=0;
	pnext1=group_definition[0];
	do{
	  strcpy(entry[nentry],strtok_r(pnext1," \t,:",&pnext1));
	  nentry++;
	}while(strlen(pnext1));
      
      //make FROM phrase
      for(i=0;i<nentry;i++){
	strcat(from_phrase,",");
	strcat(from_phrase,entry[i]);
      };
      
      //make WHERE phrase
      strcat(where_phrase," and ");
      strcat(where_phrase,group_definition[1]);
      
      for(i=0;i<nentry;i++){
	sprintf(str," and %s.%s_%s = %s.%s"
		,table_name[depth],entry[i],id_extension[depth+1],entry[i],id_extension[depth+1]);
	strcat(where_phrase,str);
      };
      
      //make ORDER BY phase
      if(transition_depth>=0){
	strcpy(order_phrase,group_definition[2]);
      }
    }//end of    "if(transition_depth==depth+1)"
#ifdef LPSQLDEBUG
    cout<<"\tFrom  :"<<from_phrase<<endl
	<<"\tWhere :"<<where_phrase<<endl<<endl;
#endif
  }//end of  "for(depth=0;depth<=maxdepth;depth++)"
  
  sprintf(sqlquery
	  ,"SELECT %s.%s FROM %s WHERE %s"
	  ,table_name[maxdepth] ,field_name
	  ,from_phrase          ,where_phrase
	  );
  if(transition_depth>=0){
    strcat(sqlquery," ORDER BY ");
    strcat(sqlquery,order_phrase);
  }
  
  sprintf(limit_phrase," LIMIT %d",max);
  strcat(sqlquery,limit_phrase);
  strcat(sqlquery,";");
#ifdef LPSQLDEBUG  
  cout<<sqlquery<<endl;
#endif
  result = connection->Query(sqlquery);

#ifdef LPSQLDEBUG
  newTime = clock();
  cout<<endl<<"Successfully done.  ("<<setprecision(3)<<setiosflags(ios::fixed)<<setw(4)<<(newTime - oldTime)/(float)CLOCKS_PER_SEC<<" sec.)"<<endl
      <<"--------------------------------------"<<endl;
  ;
#endif
  return result;
}

int lpsqlwrite(Connection *connection, char *path,int id,char **value,char *constraint,int number){
  char sqlpath[1024];
  char sqlconstraint[1024];
  strcpy(sqlpath,path);
  strcpy(sqlconstraint,constraint);
  
  Result     *result;
  char       sqlquery[2048];
  
  char where_phrase[4096]="";
  char table_name[32][64];
  char field_name[64];
  char* pnext1 = sqlpath;
  char constraint_element[64][64];
  int  nconstraint = 0;
  int  maxdepth     = 0;  
  int  i;
  char id_extension[32][8]; // "id","idx"
  char *ptable_name[32];
  char *pid_extension[32];
  char *start_id_extension="id";
  char group_name[128];
  int  transition_depth;

  for(i=0;i<32;i++){
    ptable_name[i]=table_name[i];
    pid_extension[i]=id_extension[i];
  }

  //--  decode SQL path  --//
  maxdepth = lpsqldecodepath(path,start_id_extension
			     ,ptable_name,pid_extension,transition_depth
			     ,group_name,field_name);
 
  //--decode  constraint  --//
  /*
    filedname1=value1,filedname2=value2
  */
  pnext1 = sqlconstraint;
  while(strlen(pnext1)){
    strcpy(constraint_element[nconstraint],strtok_r(pnext1,",",&pnext1));
    nconstraint++;
  }    
  
#ifdef LPSQLDEBUG
  cout<<"Table\t:"<<table_name[maxdepth]<<endl;
  clock_t oldTime,newTime;
  oldTime = clock();
  cout<<"Field\t:"<<field_name<<endl<<endl;
#endif
  int j;
  //insert/update row
  for(i=0;i<number;i++){

    if(transition_depth>=0){
      sprintf(where_phrase,"id=%d AND idx=%d",id,i);
    }
    else{
      sprintf(where_phrase,"id=\"%d\"",id);
    }
    for(j=0;j<nconstraint;j++){
      strcat(where_phrase," AND ");
      strcat(where_phrase,constraint_element[j]);
    }
    sprintf(sqlquery
	    ,"SELECT %s FROM %s WHERE %s"
	    ,field_name,table_name[maxdepth] ,where_phrase
	    );
    strcat(sqlquery,";");
#ifdef LPSQLDEBUG  
    cout<<sqlquery<<endl;
#endif
    result = connection->Query(sqlquery);
    if(!result->GetRowCount()){// if not exist
      if(transition_depth>=0){
      sprintf(sqlquery
	      ,"Insert INTO %s SET id=%d,idx=%d"
	      ,table_name[maxdepth] ,id,i
	      );
      }
      else{
      sprintf(sqlquery
	      ,"Insert INTO %s SET id=%d"
	      ,table_name[maxdepth] ,id
	      );
      }
      if(strlen(constraint)){
	strcat(sqlquery,",");
	strcat(sqlquery,constraint);
      }
      strcat(sqlquery,";");
#ifdef LPSQLDEBUG
      cout<<sqlquery<<endl;
#endif
#ifndef READONLY
    connection->Query(sqlquery);
#endif
    }
    sprintf(sqlquery
	    ,"UPDATE %s SET %s=\"%s\""
	    ,table_name[maxdepth] ,field_name,value[i]
	    );
    strcat(sqlquery," WHERE ");
    strcat(sqlquery,where_phrase);
    strcat(sqlquery,";");
#ifdef LPSQLDEBUG  
    cout<<sqlquery<<endl;
#endif
#ifndef READONLY
    connection->Query(sqlquery);
#endif
  }
  
#ifdef LPSQLDEBUG
  newTime = clock();
  cout<<endl<<"Successfully done.  ("<<setprecision(3)<<setiosflags(ios::fixed)<<setw(4)<<(newTime - oldTime)/(float)CLOCKS_PER_SEC<<" sec.)"<<endl
      <<"--------------------------------------"<<endl;
  ;
#endif
  return 0;
}

